from checkov.graph.checks.checks_infra.solvers.base_solver import BaseSolver
from checkov.graph.checks.checks_infra.enums import SolverType


class BaseFilterSolver(BaseSolver):
    operator = ''

    def __init__(self, resource_types, query_attribute, query_value):
        super().__init__(SolverType.FILTER)
        self.resource_types = resource_types
        self.query_attribute = query_attribute
        self.query_value = query_value
        self.vertices = []

    def get_operation(self, *args, **kwargs):
        raise NotImplementedError()

    def _get_operation(self, *args, **kwargs):
        raise NotImplementedError()

    def run(self, graph_connector):
        raise NotImplementedError()





from checkov.graph.terraform.checks_infra.solvers.connections_solvers.base_connection_solver import BaseConnectionSolver
from checkov.graph.checks.checks_infra.enums import SolverType
from networkx.classes.digraph import DiGraph
from checkov.graph.terraform.graph_builder.graph_components.attribute_names import CustomAttributes


class ComplexConnectionSolver(BaseConnectionSolver):
    operator = ''

    def __init__(self, queries, operator):
        if queries is None:
            queries = []
        self.queries = queries
        self.operator = operator

        resource_types = []
        connected_resources_types = []
        for sub_query in self.queries:
            if sub_query.solver_type in [SolverType.CONNECTION, SolverType.COMPLEX_CONNECTION]:
                resource_types.extend(sub_query.resource_types)
                connected_resources_types.extend(sub_query.connected_resources_types)
        resource_types = list(set(resource_types))
        connected_resources_types = list(set(connected_resources_types))

        super().__init__(resource_types, connected_resources_types)

    def run(self, graph_connector: DiGraph):
        raise NotImplementedError

    @staticmethod
    def filter_duplicates(checks):
        return list({check[CustomAttributes.ID]: check for check in checks}.values())

    def filter_results(self, passed: list, failed: list):
        filters = []
        filter_queries = [sub_query for sub_query in self.queries if sub_query.solver_type == SolverType.FILTER]
        for sub_query in filter_queries:
            filters.append(sub_query.get_operation())
        if filters:
            for query_filter in filters:
                passed = list(filter(query_filter, passed))
                failed = list(filter(query_filter, failed))
        passed = self.filter_duplicates(passed)
        failed = self.filter_duplicates(failed)
        return passed, failed



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.kubernetes.base_spec_check import BaseK8Check

class ApiServerAdmissionControlEventRateLimit(BaseK8Check):
    def __init__(self):
        # CIS-1.6 1.2.10
        id = "CKV_K8S_78"
        name = "Ensure that the admission control plugin EventRateLimit is set"
        categories = [CheckCategories.KUBERNETES]
        supported_kind = ['AdmissionConfiguration']
        super().__init__(name=name, id=id, categories=categories, supported_entities=supported_kind)

    def get_resource_id(self, conf):
        if "metadata" in conf:
            if "name" in conf["metadata"]:
                return 'AdmissionConfiguration.{}'.format(conf["metadata"]["name"])
        return 'AdmissionConfiguration.spec.allowedCapabilities'

    def scan_spec_conf(self, conf):
        if "plugins" not in conf:
            return CheckResult.FAILED
        plugins = conf["plugins"]
        for plugin in plugins:
            if plugin["name"] == "EventRateLimit":
                return CheckResult.PASSED

        return CheckResult.FAILED

check = ApiServerAdmissionControlEventRateLimit()



from checkov.cloudformation.checks.resource.base_resource_value_check import BaseResourceValueCheck
from checkov.common.models.enums import CheckCategories


class RDSDeletionProtection(BaseResourceValueCheck):

    def __init__(self):
        name = "Ensure that RDS clusters have deletion protection enabled"
        id = "CKV_AWS_139"
        supported_resources = ['aws_rds_cluster']
        categories = [CheckCategories.GENERAL_SECURITY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self):
        return 'deletion_protection'


check = RDSDeletionProtection()



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceCheck


class NetworkInterfacePublicIPAddressId(BaseResourceCheck):
    def __init__(self):
        name = "Ensure that Network Interfaces don't use public IPs"
        id = "CKV_AZURE_119"
        supported_resources = ['azurerm_network_interface']
        categories = [CheckCategories.NETWORKING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        ip_configurations = conf.get('ip_configuration', [])
        for ip_configuration in ip_configurations:
            if 'public_ip_address_id' in ip_configuration:
                return CheckResult.FAILED
        return CheckResult.PASSED


check = NetworkInterfacePublicIPAddressId()



import networkx as nx

from checkov.graph.db_connectors.db_connector import DBConnector
from checkov.graph.graph_builder.graph_components.attribute_names import CustomAttributes


class NetworkxConnector(DBConnector):
    def __init__(self):
        self.graph = nx.DiGraph()

    def save_graph(self, local_graph, add_bulk_edges=False):
        return self.networkx_from_local_graph(local_graph)

    def get_reader_endpoint(self):
        return self.graph

    def get_writer_endpoint(self):
        return self.graph

    def networkx_from_local_graph(self, local_graph):
        self.graph = nx.DiGraph()
        vertices_attributes = [v.get_decoded_attribute_dict() for v in local_graph.vertices]
        vertices_to_add = [(attr[CustomAttributes.HASH], attr) for attr in vertices_attributes]
        edges_to_add = [(vertices_attributes[e.origin][CustomAttributes.HASH], vertices_attributes[e.dest][CustomAttributes.HASH], {'label': e.label}) for e in local_graph.edges]

        self.graph.add_nodes_from(vertices_to_add)
        self.graph.add_edges_from(edges_to_add)

        return self.graph



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck


class ConfigConfigurationAggregator(BaseResourceCheck):
    def __init__(self):
        name = "Ensure AWS Config is enabled in all regions"
        id = "CKV_AWS_121"
        supported_resources = ['aws_config_configuration_aggregator']
        categories = [CheckCategories.LOGGING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        """
            Looks for account_aggregation_source /  organization_aggregation_source
            at aws_config_configuration_aggregator:
            https://registry.terraform.io/providers/hashicorp/aws/latest/docs/resources/config_configuration_aggregator#account-based-aggregation
        :param conf: aws_config_configuration_aggregator configuration
        :return: <CheckResult>
        """
        self.evaluated_keys = ["account_aggregation_source", "organization_aggregation_source"]

        if "account_aggregation_source" in conf:
            all_regions_attr = conf.get("account_aggregation_source", {})[0].get("all_regions")
            if all_regions_attr:
                return CheckResult.PASSED
            return CheckResult.FAILED
        if "organization_aggregation_source" in conf:
            if conf.get("organization_aggregation_source", {})[0].get("all_regions"):
                return CheckResult.PASSED
            return CheckResult.FAILED
        return CheckResult.FAILED


check = ConfigConfigurationAggregator()



from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.any_attribute_solver import AnyResourceSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.contains_attribute_solver import ContainsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.not_contains_attribute_solver import NotContainsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.ending_with_attribute_solver import EndingWithAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.equals_attribute_solver import EqualsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.exists_attribute_solver import ExistsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.not_ending_with_attribute_solver import NotEndingWithAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.not_equals_attribute_solver import NotEqualsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.not_exists_attribute_solver import NotExistsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.not_starting_with_attribute_solver import NotStartingWithAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.starting_with_attribute_solver import StartingWithAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_solvers.within_attribute_solver import WithinAttributeSolver


from checkov.terraform.context_parsers.tf_plan import parse

simple_types = [str, int, float, bool]


def _is_simple_type(obj):
    for simple_type in simple_types:
        if isinstance(obj, simple_type) or obj == None:
            return True
    return False


def _is_list_of_simple_types(l):
    if not isinstance(l, list):
        return False
    for i in l:
        if not _is_simple_type(i):
            return False
    return True


def _is_list_of_dicts(l):
    if not isinstance(l, list):
        return False
    for i in l:
        if isinstance(i, dict):
            return True
    return False


def _hclify(obj, parent_key=None):
    ret_dict = {}
    if not isinstance(obj, dict):
        raise Exception("this method receives only dicts")
    if hasattr(obj, 'start_mark') and hasattr(obj, "end_mark"):
        obj["start_line"] = obj.start_mark.line
        obj["end_line"] = obj.end_mark.line
    for key, value in obj.items():
        if _is_simple_type(value) or _is_list_of_simple_types(value):
            if parent_key == "tags":
                ret_dict[key] = value
            else:
                ret_dict[key] = [value]

        if _is_list_of_dicts(value):
            child_list = []
            for internal_val in value:
                child_list.append(_hclify(internal_val))
            ret_dict[key] = child_list
        if isinstance(value, dict):
            child_dict = _hclify(value, key)
            if parent_key == "tags":
                ret_dict[key] = child_dict
            else:
                ret_dict[key] = [child_dict]
    return ret_dict

def _prepare_resource_block(resource):
    """
    hclify resource if pre-conditions met.
    :type: resource: dict: tf resource block
    :rtype: resource_block: dict: hclifyed if conditions met
    :rtype: prepared: boolean: whether conditions met to prepare data
    """
    resource_block = {}
    resource_block[resource['type']] = {}
    prepared = False
    mode = ""
    if 'mode' in resource:
        mode = resource.get("mode")
    # Rare cases where data block appears in resources with same name as resource block and only partial values
    # and where *_module resources don't have values field
    if mode == "managed" and 'values' in resource:
        resource_block[resource['type']][resource.get("name", "default")] = _hclify(resource['values'])
        prepared = True
    return resource_block, prepared

def _find_child_modules(child_modules):
    """
    Find all child modules if any. Including any amount of nested child modules.
    :type: child_modules: list of tf child_module objects
    :rtype: resource_blocks: list of hcl resources
    """
    resource_blocks = []
    for child_module in child_modules:
        if child_module.get("child_modules",[]):
            nested_child_modules = child_module.get("child_modules",[])
            nested_blocks = _find_child_modules(nested_child_modules)
            for resource in nested_blocks:
                resource_blocks.append(resource)
        for resource in child_module.get("resources", []):
            resource_block, prepared = _prepare_resource_block(resource)
            if prepared is True:
                resource_blocks.append(resource_block)
    return resource_blocks


def parse_tf_plan(tf_plan_file):
    """
    :type tf_plan_file: str - path to plan file
    :rtype: tf_definition dictionary
    """
    tf_defintions = {}
    tf_defintions[tf_plan_file] = {}
    tf_defintions[tf_plan_file]['resource'] = []
    template, template_lines = parse(tf_plan_file)
    if not template:
        return None, None
    for resource in template.get('planned_values', {}).get("root_module", {}).get("resources", []):
        resource_block, prepared = _prepare_resource_block(resource)
        if prepared is True:
            tf_defintions[tf_plan_file]['resource'].append(resource_block)
    child_modules = template.get('planned_values', {}).get("root_module", {}).get("child_modules",[])
    # Terraform supports modules within modules so we need to search
    # in nested modules to find all resource blocks
    resource_blocks = _find_child_modules(child_modules)
    for resource in resource_blocks:
        tf_defintions[tf_plan_file]['resource'].append(resource)
    return tf_defintions, template_lines



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceCheck


class StorageBlobServiceContainerPrivateAccess(BaseResourceCheck):
    def __init__(self):
        name = "Ensure that 'Public access level' is set to Private for blob containers"
        id = "CKV_AZURE_34"
        supported_resources = ['azurerm_storage_container']
        categories = [CheckCategories.NETWORKING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if 'container_access_type' in conf and conf['container_access_type'][0] != 'private':
            return CheckResult.FAILED
        return CheckResult.PASSED


check = StorageBlobServiceContainerPrivateAccess()



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.kubernetes.base_spec_check import BaseK8Check

class EtcdClientCertAuth(BaseK8Check):
    def __init__(self):
        # CIS-1.6 2.2
        id = "CKV_K8S_117"
        name = "Ensure that the --client-cert-auth argument is set to true"
        categories = [CheckCategories.KUBERNETES]
        supported_entities = ['containers']
        super().__init__(name=name, id=id, categories=categories, supported_entities=supported_entities)

    def get_resource_id(self, conf):
        return f'{conf["parent"]} - {conf["name"]}'

    def scan_spec_conf(self, conf):
        if "etcd" in conf.get("command", []) and "--client-cert-auth=true" not in conf.get("command", []):
            return CheckResult.FAILED
           
        return CheckResult.PASSED

check = EtcdClientCertAuth()


import logging
from abc import abstractmethod

from checkov.common.util.type_forcers import force_list
from checkov.common.models.enums import CheckResult
from checkov.common.multi_signature import MultiSignatureMeta, multi_signature


class BaseCheck(metaclass=MultiSignatureMeta):
    id = ""
    name = ""
    categories = []
    supported_entities = []

    def __init__(self, name, id, categories, supported_entities, block_type):
        self.name = name
        self.id = id
        self.categories = categories
        self.block_type = block_type
        self.supported_entities = supported_entities
        self.logger = logging.getLogger("{}".format(self.__module__))
        self.evaluated_keys = []

    def run(self, scanned_file, entity_configuration, entity_name, entity_type, skip_info):
        check_result = {}
        if skip_info:
            check_result['result'] = CheckResult.SKIPPED
            check_result['suppress_comment'] = skip_info['suppress_comment']
            message = "File {}, {} \"{}.{}\" check \"{}\" Result: {}, Suppression comment: {} ".format(
                scanned_file,
                self.block_type,
                entity_type,
                entity_name,
                self.name,
                check_result,
                check_result['suppress_comment']
            )
            self.logger.debug(message)
        else:
            try:
                self.evaluated_keys = []
                check_result['result'] = self.scan_entity_conf(entity_configuration, entity_type)
                check_result['evaluated_keys'] = self.get_evaluated_keys()
                message = "File {}, {}  \"{}.{}\" check \"{}\" Result: {} ".format(
                    scanned_file,
                    self.block_type,
                    entity_type,
                    entity_name,
                    self.name,
                    check_result
                )
                self.logger.debug(message)

            except Exception as e:
                self.logger.error(
                    "Failed to run check: {} for configuration: {} at file: {}".format(
                        self.name,
                        str(entity_configuration),
                        scanned_file
                    )
                )
                raise e
        return check_result

    @multi_signature()
    @abstractmethod
    def scan_entity_conf(self, conf, entity_type):
        raise NotImplementedError()

    @classmethod
    @scan_entity_conf.add_signature(args=["self", "conf"])
    def _scan_entity_conf_self_conf(cls, wrapped):
        def wrapper(self, conf, entity_type=None):
            # keep default argument for entity_type so old code, that doesn't set it, will work.
            return wrapped(self, conf)

        return wrapper

    def get_evaluated_keys(self):
        """
        Retrieves the evaluated keys for the run's report. Child classes override the function and return the `expected_keys` instead.
        :return: List of the evaluated keys, as JSONPath syntax paths of the checked attributes
        """
        return force_list(self.evaluated_keys)



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.kubernetes.base_spec_check import BaseK8Check


class ApiServerTlsCertAndKey(BaseK8Check):
    def __init__(self):
        # CIS-1.6 1.2.30
        id = "CKV_K8S_100"
        name = "Ensure that the --tls-cert-file and --tls-private-key-file arguments are set as appropriate"
        categories = [CheckCategories.KUBERNETES]
        supported_entities = ['containers']
        super().__init__(name=name, id=id, categories=categories, supported_entities=supported_entities)

    def get_resource_id(self, conf):
        return f'{conf["parent"]} - {conf["name"]}'

    def scan_spec_conf(self, conf):
        if "command" in conf and conf["command"] is not None:
            if "kube-apiserver" in conf["command"]:
                hasCertCommand = False
                hasKeyCommand = False
                for command in conf["command"]:
                    if command.startswith("--tls-cert-file"):
                        hasCertCommand = True
                    elif command.startswith("--tls-private-key-file"):
                        hasKeyCommand = True
                return CheckResult.PASSED if hasCertCommand and hasKeyCommand else CheckResult.FAILED

        return CheckResult.PASSED


check = ApiServerTlsCertAndKey()



import base64
import builtins
import concurrent.futures
import hashlib
import io
import json
import os
import pickle  # nosec
import re

from checkov.graph.terraform.graph_builder.graph_components.block_types import BlockType

from checkov.graph.terraform.graph_builder.graph_components.attribute_names import CustomAttributes

BLOCK_TYPES_STRINGS = ['var', 'local', 'module', 'data']
FUNC_CALL_PREFIX_PATTERN = r'([.a-zA-Z]+)\('
INTERPOLATION_PATTERN = '[${}]'
INTERPOLATION_EXPR = r'\$\{([^\}]*)\}'
BRACKETS_PATTERN = r'[\[\]\)\(\{\}]+'
INDEX_PATTERN = r'\[([0-9]+)\]'
IDENTIFIER_PATTERN = r'^[^\d\W]\w*\Z'
MAP_ATTRIBUTE_PATTERN = r'\[\"([^\d\W]\w*)\"\]'


class VertexReference:
    def __init__(self, block_type, sub_parts, origin_value):
        self.block_type = block_type_str_to_enum(block_type) if type(block_type) is str else block_type
        self.sub_parts = sub_parts
        self.origin_value = origin_value

    def __eq__(self, other):
        if not isinstance(other, VertexReference):
            return NotImplemented
        return self.block_type == other.block_type and self.sub_parts == other.sub_parts and self.origin_value == other.origin_value

    def __str__(self):
        return f"{self.block_type} sub_parts = {self.sub_parts}, origin = {self.origin_value}"


def get_vertices_references(str_value, aliases, resources_types):
    vertices_references = []
    words_in_str_value = str_value.split()

    for word in words_in_str_value:
        if word.startswith('.') or word.startswith('/.'):
            # check if word is a relative path
            continue

        interpolations = re.split(INTERPOLATION_EXPR, word)
        for interpolation_content in interpolations:
            for w in interpolation_content.split(','):
                word_sub_parts = w.split('.')
                if len(word_sub_parts) <= 1 or word_sub_parts[0].isnumeric():
                    # if the word doesn't contain a '.' char, or if the first part before the dot is a number
                    continue

                suspected_block_type = word_sub_parts[0]
                if suspected_block_type in BLOCK_TYPES_STRINGS:
                    # matching cases like 'var.x'
                    vertex_reference = VertexReference(block_type=suspected_block_type, sub_parts=word_sub_parts[1:], origin_value=w)
                    if vertex_reference not in vertices_references:
                        vertices_references.append(vertex_reference)
                    continue

                vertex_reference = get_vertex_reference_from_alias(suspected_block_type, aliases, word_sub_parts)
                if vertex_reference and vertex_reference not in vertices_references:
                    vertex_reference.origin_value = w
                    # matching cases where the word is referring an alias
                    vertices_references.append(vertex_reference)
                    continue

                # matching cases like 'aws_vpc.main'
                if word_sub_parts[0] in resources_types:
                    block_name = word_sub_parts[0] + '.' + word_sub_parts[1]
                    word_sub_parts = [block_name] + word_sub_parts[2:]
                    vertex_reference = VertexReference(block_type=BlockType.RESOURCE, sub_parts=word_sub_parts,
                                                       origin_value=w)
                    if vertex_reference not in vertices_references:
                        vertices_references.append(vertex_reference)

    return vertices_references


def block_type_str_to_enum(block_type_str):
    if block_type_str == 'var':
        return BlockType.VARIABLE
    if block_type_str == 'local':
        return BlockType.LOCALS
    try:
        return BlockType(block_type_str)
    except Exception:
        return None


def block_type_enum_to_str(block_type):
    if block_type == BlockType.VARIABLE:
        return 'var'
    if block_type == BlockType.LOCALS:
        return 'local'
    return block_type


def get_vertex_reference_from_alias(block_type_str, aliases, val):
    block_type = ''
    if block_type_str in aliases:
        block_type = aliases[block_type_str][CustomAttributes.BLOCK_TYPE]
    aliased_provider = ".".join(val)
    if aliased_provider in aliases:
        block_type = aliases[aliased_provider][CustomAttributes.BLOCK_TYPE]
    if block_type:
        return VertexReference(block_type=block_type, sub_parts=val,
                               origin_value='')
    return None


def remove_function_calls_from_str(str_value):
    # remove start of function calls:: 'length(aws_vpc.main) > 0 ? aws_vpc.main[0].cidr_block : ${var.x}' --> 'aws_vpc.main) > 0 ? aws_vpc.main[0].cidr_block : ${var.x}'
    str_value = re.sub(FUNC_CALL_PREFIX_PATTERN, '', str_value)
    # remove ')'
    return re.sub(r'[)]+', '', str_value)


def remove_index_pattern_from_str(str_value):
    str_value = re.sub(INDEX_PATTERN, '', str_value)
    str_value = str_value.replace("[", " [ ")
    str_value = str_value.replace("]", " ] ")
    return str_value


def remove_interpolation(str_value, replace_str=' '):
    return re.sub(INTERPOLATION_PATTERN, replace_str, str_value)


def replace_map_attribute_access_with_dot(str_value):
    split_by_identifiers = re.split(MAP_ATTRIBUTE_PATTERN, str_value)
    new_split = []
    for split_part in split_by_identifiers:
        if split_part.startswith('.'):
            split_part = split_part[1:]
        if split_part.endswith('.'):
            split_part = split_part[:-1]
        new_split.append(split_part)

    return '.'.join(new_split)


DEFAULT_CLEANUP_FUNCTIONS = [remove_function_calls_from_str, remove_index_pattern_from_str, replace_map_attribute_access_with_dot, remove_interpolation]


def get_referenced_vertices_in_value(value, aliases, resources_types, cleanup_functions=DEFAULT_CLEANUP_FUNCTIONS):
    references_vertices = []
    value_type = type(value)

    if value_type is list:
        for sub_value in value:
            references_vertices += get_referenced_vertices_in_value(sub_value, aliases, resources_types, cleanup_functions)

    if value_type is dict:
        for sub_key in value:
            references_vertices += get_referenced_vertices_in_value(value[sub_key], aliases, resources_types, cleanup_functions)

    if value_type is str:
        if cleanup_functions:
            for func in cleanup_functions:
                value = func.__call__(value)
        references_vertices = get_vertices_references(value, aliases, resources_types)

    return references_vertices


def encode_graph_property_value(value):
    if isinstance(value, bool):
        # Encode boolean into Terraform's lower case convention
        value = str(value).lower()
    elif isinstance(value, (float, int)):
        value = str(value)
    return json.dumps(value, cls=PythonObjectEncoder, indent=4)


def decode_graph_property_value(value, leave_str=False):
    try:
        if type(value) is str and value.startswith('"') and value.endswith('"'):
            value = value[1:-1]
        if not leave_str:
            if value.isnumeric():
                value = int(value)
            value = json.loads(value, object_hook=as_python_object)
        return value
    except (ValueError, AttributeError):
        try:
            if type(value) in [str, bytes, bytearray]:
                value = json.loads(value)
            return value
        except ValueError:
            pass

    finally:
        return value


class RestrictedUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        # Only allow safe classes from builtins.
        if module == "builtins" or module.startswith("lark."):
            return getattr(builtins, name)
        # Forbid everything else.
        raise pickle.UnpicklingError("global '%s.%s' is forbidden" %
                                     (module, name))

def restricted_loads(s):
    """Helper function analogous to pickle.loads()."""
    return RestrictedUnpickler(io.BytesIO(s)).load()


class PythonObjectEncoder(json.JSONEncoder):
    def default(self, obj):
        return {'_python_object':
                base64.b64encode(pickle.dumps(obj)).decode('utf-8') }


def as_python_object(dct):
    if '_python_object' in dct:
        return restricted_loads(base64.b64decode(dct['_python_object']))
    return dct


def calculate_hash(data):
    encoded_attributes = encode_graph_property_value(data)
    sha256 = hashlib.sha256()
    sha256.update(repr(encoded_attributes).encode('utf-8'))

    return sha256.hexdigest()


def run_function_multithreaded(func, data, max_group_size, num_of_workers=None):
    groups_of_data = [data[i:i + max_group_size] for i in
                      range(0, len(data), max_group_size)]
    if not num_of_workers:
        num_of_workers = len(groups_of_data)
    if num_of_workers > 0:
        with concurrent.futures.ThreadPoolExecutor(max_workers=num_of_workers) as executor:
            futures = {executor.submit(func, data_group): data_group for data_group in groups_of_data}
            wait_result = concurrent.futures.wait(futures)
            if wait_result.not_done:
                raise Exception(f'failed to perform {func.__name__}')
            for future in futures:
                try:
                    future.result()
                except Exception as e:
                    raise e


def update_dictionary_attribute(config, key_to_update, new_value):
    key_parts = key_to_update.split('.')
    if type(config) is dict:
        if config.get(key_parts[0]) is not None:
            key = key_parts[0]
            if len(key_parts) == 1:
                if isinstance(config[key], list) and not isinstance(new_value, list):
                    new_value = [new_value]
                config[key] = new_value
                return config
            else:
                config[key] = update_dictionary_attribute(config[key], '.'.join(key_parts[1:]), new_value)
        else:
            for key in config:
                config[key] = update_dictionary_attribute(config[key], key_to_update, new_value)
    if type(config) is list:
        for i in range(len(config)):
            config[i] = update_dictionary_attribute(config[i], key_to_update, new_value)

    return config

def join_trimmed_strings(char_to_join, str_lst, num_to_trim):
    return char_to_join.join(str_lst[:len(str_lst) - num_to_trim])


def filter_sub_keys(key_list):
    filtered_key_list = []
    for key in key_list:
        if not any(other_key != key and other_key.startswith(key) for other_key in key_list):
            filtered_key_list.append(key)
    return filtered_key_list

def generate_possible_strings_from_wildcards(origin_string, max_entries=10):
    max_entries = int(os.environ.get("MAX_WILDCARD_ARR_SIZE", max_entries))
    generated_strings = [origin_string]
    if not origin_string:
        return []
    if '*' not in origin_string:
        return generated_strings

    locations_of_wildcards = []
    for i, char in enumerate(origin_string):
        if char == '*':
            locations_of_wildcards.append(i)
    locations_of_wildcards.reverse()

    for wildcard_index in locations_of_wildcards:
        new_generated_strings = []
        for s in generated_strings:
            before_wildcard = s[:wildcard_index]
            after_wildcard = s[wildcard_index+1:]
            for i in range(max_entries):
                new_generated_strings.append(before_wildcard + str(i) + after_wildcard)
        generated_strings = new_generated_strings

    # if origin_string == "ingress.*.cidr_blocks", check for "ingress.cidr_blocks" too
    generated_strings.append(''.join(origin_string.split('.*')))
    return generated_strings


def extend_referenced_vertices_with_tf_vars(referenced_vertices):
    tfvars_to_add = []
    for vertex_ref in referenced_vertices:
        if vertex_ref.block_type == BlockType.VARIABLE:
            tfvars_to_add.append(VertexReference(block_type=BlockType.TF_VARIABLE, sub_parts=vertex_ref.sub_parts, origin_value=vertex_ref.origin_value))
    referenced_vertices.extend(tfvars_to_add)



from checkov.graph.terraform.checks_infra.solvers.connections_solvers.base_connection_solver import BaseConnectionSolver
from networkx.classes.digraph import DiGraph
from networkx import edge_dfs
from checkov.graph.graph_builder.graph_components.attribute_names import CustomAttributes
from checkov.graph.terraform.graph_builder.graph_components.block_types import BlockType


class ConnectionExistsSolver(BaseConnectionSolver):
    operator = 'exists'

    def __init__(self, resource_types, connected_resources_types, vertices_under_resource_types=None, vertices_under_connected_resources_types=None):
        super().__init__(resource_types, connected_resources_types, vertices_under_resource_types, vertices_under_connected_resources_types)

    def run(self, graph_connector: DiGraph):
        passed, failed = [], []
        for u, v in edge_dfs(graph_connector):
            origin_attributes = graph_connector.nodes(data=True)[u]
            destination_attributes = graph_connector.nodes(data=True)[v]
            origin_type = origin_attributes.get(CustomAttributes.RESOURCE_TYPE)
            destination_type = destination_attributes.get(CustomAttributes.RESOURCE_TYPE)
            if self.is_associated_edge(origin_type, destination_type):
                passed.extend([origin_attributes, destination_attributes])
            destination_block_type = destination_attributes.get(CustomAttributes.BLOCK_TYPE)
            if destination_block_type == BlockType.OUTPUT.value:
                try:
                    output_edges = graph_connector.edges(v, data=True)
                    _, output_destination, _ = next(iter(output_edges))
                    output_destination = graph_connector.nodes(data=True)[output_destination]
                    output_destination_type = output_destination.get(CustomAttributes.RESOURCE_TYPE)
                    if self.is_associated_edge(origin_type, output_destination_type):
                        passed.extend([origin_attributes, output_destination])
                except StopIteration:
                    continue
        for v, v_data in graph_connector.nodes(data=True):
            v_type = v_data.get(CustomAttributes.RESOURCE_TYPE)
            if self.is_associated_vertex(v_type):
                v_degree = graph_connector.degree(v)
                if v_degree == 0:
                    failed.append(v_data)
                else:
                    is_associated = False
                    for s, t, _ in graph_connector.edges(v, data=True):
                        s_attributes = graph_connector.nodes(data=True)[s]
                        t_attributes = graph_connector.nodes(data=True)[t]
                        s_type = s_attributes.get(CustomAttributes.RESOURCE_TYPE)
                        t_type = t_attributes.get(CustomAttributes.RESOURCE_TYPE)
                        if self.is_associated_edge(s_type, t_type):
                            is_associated = True
                    if not is_associated:
                        failed.append(v_data)
        return passed, failed



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck


class RDSClusterEncrypted(BaseResourceCheck):
    def __init__(self):
        name = "Ensure that RDS global clusters are encrypted"
        id = "CKV_AWS_149"
        supported_resources = ['aws_rds_global_cluster']
        categories = [CheckCategories.ENCRYPTION]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        """
            Looks for storage_encrypted at aws_rds_global_cluster:
            https://registry.terraform.io/providers/hashicorp/aws/latest/docs/resources/rds_global_cluster
        :param conf: aws_rds_global_cluster configuration
        :return: <CheckResult>
        """
        self.evaluated_keys = 'aws_rds_global_cluster'
        if "source_db_cluster_identifier" in conf.keys():
            return CheckResult.UNKNOWN
        if "storage_encrypted" in conf.keys():
            if conf["storage_encrypted"][0]:
                return CheckResult.PASSED
            return CheckResult.FAILED
        return CheckResult.FAILED


check = RDSClusterEncrypted()



import os
import re

MODULE_DEPENDENCY_PATTERN_IN_PATH = r'\[.+\#.+\]'


def is_local_path(root_dir, source):
    # https://www.terraform.io/docs/modules/sources.html#local-paths
    return source.startswith('./') or \
           source.startswith('/./') or \
           source.startswith('../') or \
           source in os.listdir(root_dir)


def remove_module_dependency_in_path(path):
    """
    :param path: path that looks like "dir/main.tf[other_dir/x.tf#0]
    :return: separated path from module dependency: dir/main.tf, other_dir/x.tf
    """
    module_dependency = re.findall(MODULE_DEPENDENCY_PATTERN_IN_PATH, path)
    if re.findall(MODULE_DEPENDENCY_PATTERN_IN_PATH, path):
        path = re.sub(MODULE_DEPENDENCY_PATTERN_IN_PATH, '', path)
    return path, extract_module_dependency_path(module_dependency)


def extract_module_dependency_path(module_dependency):
    """
    :param module_dependency: a list looking like ['[path_to_module.tf#0]']
    :return: the path without enclosing array and index: 'path_to_module.tf'
    """
    if not module_dependency:
        return ''
    if type(module_dependency) is list and len(module_dependency) > 0:
        module_dependency = module_dependency[0]
    return module_dependency[1:-1].split('#')[0]



from checkov.graph.checks.checks_infra.base_check import BaseGraphCheck


class BaseGraphCheckParser:
    def parse_raw_check(self, raw_check, **kwargs) -> BaseGraphCheck:
        raise NotImplementedError



import copy
import logging
import os
from typing import Dict

import dpath.util

from checkov.common.output.record import Record
from checkov.common.output.report import Report
from checkov.common.runners.base_runner import BaseRunner
from checkov.common.variables.context import EvaluationContext
from checkov.graph.db_connectors.networkx.networkx_db_connector import NetworkxConnector
from checkov.graph.graph_record import GraphRecord
from checkov.graph.terraform.checks_infra.nx_checks_parser import NXGraphCheckParser
from checkov.graph.terraform.checks_infra.registry import Registry
from checkov.graph.terraform.graph_builder.graph_components.attribute_names import CustomAttributes
from checkov.graph.terraform.graph_builder.graph_to_tf_definitions import convert_graph_vertices_to_tf_definitions
from checkov.graph.terraform.graph_builder.local_graph import LocalGraph
from checkov.graph.terraform.graph_manager import GraphManager
from checkov.graph.terraform.parser import TerraformGraphParser
from checkov.runner_filter import RunnerFilter
from checkov.terraform.checks.resource.registry import resource_registry
from checkov.terraform.runner import Runner as TerraformRunner

LOG_LEVEL = os.getenv('LOG_LEVEL', 'WARNING').upper()
logging.basicConfig(level=LOG_LEVEL)


class Runner(BaseRunner):
    check_type = "terraform"

    def __init__(self, parser=TerraformGraphParser(), db_connector=NetworkxConnector(), external_registries=None,
                 source="Terraform", graph_class=LocalGraph, graph_manager=None):
        self.external_registries = [] if external_registries is None else external_registries
        self.graph_class = graph_class
        self.parser = parser
        self.tf_definitions = None
        self.definitions_context = None
        self.breadcrumbs = None
        self.evaluations_context: Dict[str, Dict[str, EvaluationContext]] = {}
        self.graph_manager = graph_manager if graph_manager is not None else GraphManager(source=source, db_connector=db_connector)
        self.tf_runner = TerraformRunner()
        self.graph = None

    def set_external_data(self, tf_definitions: dict, definitions_context: dict, breadcrumbs: dict):
        self.tf_definitions = tf_definitions
        self.definitions_context = definitions_context
        self.breadcrumbs = breadcrumbs

    def run(self, root_folder, external_checks_dir=None, files=None, runner_filter=RunnerFilter(),
            collect_skip_comments=True):
        report = Report(self.check_type)
        parsing_errors = {}
        if external_checks_dir:
            for directory in external_checks_dir:
                resource_registry.load_external_checks(directory, runner_filter)
        if self.definitions_context is None or self.tf_definitions is None or self.breadcrumbs is None:
            self.tf_definitions = {}
            logging.info("Scanning root folder and producing fresh tf_definitions and context")
            if root_folder:
                root_folder = os.path.abspath(root_folder)

                local_graph, tf_definitions = self.graph_manager.build_graph_from_source_directory(root_folder, self.graph_class)
            elif files:
                files = [os.path.abspath(file) for file in files]
                root_folder = os.path.split(os.path.commonprefix(files))[0]
                self.parser.evaluate_variables = False
                for file in files:
                    if file.endswith(".tf"):
                        file_parsing_errors = {}
                        parse_result = self.parser.parse_file(file=file, parsing_errors=file_parsing_errors)
                        if parse_result is not None:
                            self.tf_runner.tf_definitions[file] = parse_result
                        if file_parsing_errors:
                            parsing_errors.update(file_parsing_errors)
                            continue
                local_graph = self.graph_manager.build_graph_from_tf_definitions(self.tf_runner.tf_definitions)
            else:
                raise Exception("Root directory was not specified, files were not specified")

            self.graph = self.graph_manager.save_graph(local_graph)
            self.tf_runner.tf_definitions, self.breadcrumbs = convert_graph_vertices_to_tf_definitions(local_graph.vertices, root_folder)
        else:
            logging.info(f"Scanning root folder using existing tf_definitions")
            self.tf_runner.tf_definitions = self.tf_definitions

        self.tf_runner.check_tf_definition(report, root_folder, runner_filter, collect_skip_comments, self.definitions_context)
        self.definitions_context = self.tf_runner.definitions_context

        report.add_parsing_errors(parsing_errors.keys())

        graph_report = self.get_graph_checks_report(root_folder, self.breadcrumbs)
        merge_reports(report, graph_report)

        return report

    def get_graph_checks_report(self, root_folder, breadcrumbs):
        registry = Registry(parser=NXGraphCheckParser())
        report = Report(self.check_type)
        checks_results = {}
        for r in self.external_registries + [registry]:
            r.load_checks()
            registry_results = r.run_checks(self.graph_manager.get_reader_traversal())
            checks_results = {**checks_results, **registry_results}

        for check_id, check_results in checks_results.items():
            for check_result in check_results:
                entity = check_result['entity']
                entity_context, entity_evaluations = self.get_entity_context_and_evaluations(entity)
                if entity_context:
                    full_file_path = entity[CustomAttributes.FILE_PATH]
                    copy_of_check_result = copy.deepcopy(check_result)
                    copy_of_check_result['entity'] = entity.get(CustomAttributes.CONFIG)
                    record = Record(check_id=check_id,
                                    check_name=check_id,
                                    check_result=copy_of_check_result,
                                    code_block=entity_context.get('code_lines'),
                                    file_path=f"/{os.path.relpath(full_file_path, root_folder)}",
                                    file_line_range=[entity_context.get('start_line'), entity_context.get('end_line')],
                                    resource=".".join(entity_context['definition_path']),
                                    evaluations=entity_evaluations,
                                    check_class=check_id,
                                    file_abs_path=os.path.abspath(full_file_path))
                    breadcrumb = breadcrumbs.get(record.file_path, {}).get(record.resource)
                    if breadcrumb:
                        record = GraphRecord(record, breadcrumb)

                    report.add_record(record=record)
        return report

    def get_entity_context_and_evaluations(self, entity):
        entity_evaluations = None
        block_type = entity[CustomAttributes.BLOCK_TYPE]
        full_file_path = entity[CustomAttributes.FILE_PATH]
        definition_path = entity[CustomAttributes.BLOCK_NAME].split('.')
        entity_context = None
        entity_context_path = [block_type] + definition_path
        if dpath.search(self.definitions_context.get(full_file_path), entity_context_path):
            entity_context = dpath.get(self.definitions_context[full_file_path], entity_context_path)
            entity_context['definition_path'] = definition_path
        return entity_context, entity_evaluations


def merge_reports(base_report, report_to_merge):
    base_report.passed_checks.extend(report_to_merge.passed_checks)
    base_report.failed_checks.extend(report_to_merge.failed_checks)
    base_report.skipped_checks.extend(report_to_merge.skipped_checks)
    base_report.parsing_errors.extend(report_to_merge.parsing_errors)



from checkov.graph.checks.checks_infra.enums import SolverType
from checkov.graph.checks.checks_infra.solvers.base_solver import BaseSolver


class BaseComplexSolver(BaseSolver):
    operator = ''

    def __init__(self, queries, resource_types):
        if queries is None:
            queries = []
        self.queries = queries
        self.resource_types = resource_types
        super().__init__(SolverType.COMPLEX)

    def get_operation(self, *args, **kwargs):
        predicates = []
        for i, query in enumerate(self.queries):
            predicates.append(query.get_operation(args[0]))
        return self._get_operation(*predicates)

    def _get_operation(self, *args, **kwargs):
        raise NotImplementedError()

    def _get_negative_op(self, *args):
        return not self._get_operation(args)

    def run(self, graph_connector):
        all_vertices_resource_types = [data for _, data in graph_connector.nodes(data=True) if
                                       self.resource_type_pred(data)]
        passed_vertices = [data for data in all_vertices_resource_types if self.get_operation(data)]
        failed_vertices = [resource for resource in all_vertices_resource_types if resource not in passed_vertices]
        return passed_vertices, failed_vertices

    def resource_type_pred(self, v):
        return len(self.resource_types) == 0 or v.get('resource_type') in self.resource_types



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck


class VMScaleSetsAutoOSImagePatchingEnabled(BaseResourceCheck):
    def __init__(self):
        name = "Ensure that automatic OS image patching is enabled for Virtual Machine Scale Sets"
        id = "CKV_AZURE_95"
        supported_resources = ['azurerm_virtual_machine_scale_set']
        categories = [CheckCategories.GENERAL_SECURITY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if 'automatic_os_upgrade' in conf and conf['automatic_os_upgrade'][0]:
            if 'os_profile_windows_config' in conf and conf['os_profile_windows_config'][0]:
                os_profile_windows_config = conf['os_profile_windows_config'][0]
                if 'enable_automatic_upgrades' in os_profile_windows_config \
                        and os_profile_windows_config['enable_automatic_upgrades'][0]:
                    return CheckResult.PASSED
        return CheckResult.FAILED


check = VMScaleSetsAutoOSImagePatchingEnabled()



import logging
import os
from copy import deepcopy

from checkov.graph.terraform.graph_builder.graph_components.block_types import BlockType
from checkov.terraform.parser import Parser
from checkov.graph.terraform.graph_builder.graph_components.module import Module
from checkov.common.util.consts import DEFAULT_EXTERNAL_MODULES_DIR

from checkov.graph.terraform.graph_builder.utils import remove_module_dependency_in_path

external_modules_download_path = os.environ.get('EXTERNAL_MODULES_DIR', DEFAULT_EXTERNAL_MODULES_DIR)


class TerraformGraphParser(Parser):
    def __init__(self, module_class=Module):
        super().__init__()
        self.module_class = module_class

    def parse_hcl_module(self, source_dir, source):
        tf_definitions = {}
        parsing_errors = {}
        download_external_modules = os.environ.get('DOWNLOAD_EXTERNAL_MODULES', 'false').lower() == 'true'
        self.parse_directory(directory=source_dir, out_definitions=tf_definitions, out_evaluations_context={},
                             out_parsing_errors=parsing_errors,
                             download_external_modules=download_external_modules, evaluate_variables=False,
                             external_modules_download_path=external_modules_download_path)
        tf_definitions = TerraformGraphParser._hcl_boolean_types_to_boolean(tf_definitions)
        return self.parse_hcl_module_from_tf_definitions(tf_definitions, source_dir, source)

    def parse_hcl_module_from_tf_definitions(self, tf_definitions, source_dir, source):
        module = self.get_new_module(source_dir)
        self.add_tfvars(module, source)
        module_dependency_map, tf_definitions = self.get_module_dependency_map(tf_definitions)
        copy_of_tf_definitions = deepcopy(tf_definitions)
        for file_path in copy_of_tf_definitions:
            blocks = copy_of_tf_definitions.get(file_path)
            for block_type in blocks:
                try:
                    module.add_blocks(block_type, blocks[block_type], file_path, source)
                except Exception as e:
                    logging.error(f'Failed to add block {blocks[block_type]}. Error:')
                    logging.error(e, exc_info=True)
        return module, module_dependency_map, tf_definitions

    @staticmethod
    def _hcl_boolean_types_to_boolean(conf: dict) -> dict:
        sorted_keys = list(conf.keys())
        if len(conf.keys()) > 0 and all(isinstance(x, type(list(conf.keys())[0])) for x in conf.keys()):
            sorted_keys = sorted(filter(lambda x: x is not None, conf.keys()))
        # Create a new dict where the keys are sorted alphabetically
        sorted_conf = {key: conf[key] for key in sorted_keys}
        for attribute, values in sorted_conf.items():
            if attribute is 'alias':
                continue
            if isinstance(values, list):
                sorted_conf[attribute] = TerraformGraphParser._hcl_boolean_types_to_boolean_lst(values)
            elif isinstance(values, dict):
                sorted_conf[attribute] = TerraformGraphParser._hcl_boolean_types_to_boolean(conf[attribute])
            elif isinstance(values, str) and values in ('true', 'false'):
                sorted_conf[attribute] = True if values == 'true' else False
        return sorted_conf

    @staticmethod
    def _hcl_boolean_types_to_boolean_lst(values: list) -> list:
        for i in range(len(values)):
            val = values[i]
            if isinstance(val, dict):
                values[i] = TerraformGraphParser._hcl_boolean_types_to_boolean(val)
            elif isinstance(val, list):
                values[i] = TerraformGraphParser._hcl_boolean_types_to_boolean_lst(val)
            elif isinstance(val, str):
                if val == 'true':
                    values[i] = True
                elif val == 'false':
                    values[i] = False
        str_values_in_lst = [val for val in values if isinstance(val, str)]
        str_values_in_lst.sort()
        result_values = [val for val in values if not isinstance(val, str)]
        result_values.extend(str_values_in_lst)
        return result_values

    @staticmethod
    def get_module_dependency_map(tf_definitions):
        """
        :param tf_definitions, with paths in format 'dir/main.tf[module_dir/main.tf#0]'
        :return module_dependency_map: mapping between directories and the location of its module definition:
                {'dir': 'module_dir/main.tf'}
        :return tf_definitions: with paths in format 'dir/main.tf'
        """
        module_dependency_map = {}
        copy_of_tf_definitions = {}
        for file_path in tf_definitions.keys():
            path, module_dependency = remove_module_dependency_in_path(file_path)
            dir_name = os.path.dirname(path)
            if not module_dependency_map.get(dir_name):
                module_dependency_map[dir_name] = module_dependency
            copy_of_tf_definitions[path] = deepcopy(tf_definitions[file_path])
        return module_dependency_map, copy_of_tf_definitions

    @staticmethod
    def get_new_module(source_dir):
        return Module(source_dir, encode=False)

    def add_tfvars(self, module, source):
        if not self.var_value_and_file_map:
            return
        for var_name, (default, path) in self.var_value_and_file_map.items():
            if ".tfvars" in path:
                block = {var_name: {"default": default}}
                module.add_blocks(BlockType.TF_VARIABLE, block, path, source)
